
' Virtual display version...

Import mojo

' Use virtual display system -- SetVirtualDisplay 800, 534 added to OnCreate, all
' DeviceWidth/Height calls changed to VDeviceWidth/Height and UpdateVirtualDisplay
' and ScaleVirtualDisplay added to start of OnUpdate and OnRender respectively.

' Two-minute job to convert, literally!

Import autofit

Const numdots:Int = 100 ' Number of dots drawn on the ground...

' App class...

Class Game Extends App

	Field useframe:Image			' Frame to be used for drawing, selected according to rotation...

	' Car sprite by Reiner "Tiles" Prokein -- see license at URL below before redistributing!
	
	' http://reinerstileset.4players.de/readmeE.html
	
	Field carstraight:Image [32]		' Car frames (wheels straight)
	Field carleft:Image	 [32]		' Car frames (wheels turned left)
	Field carright:Image	 [32]		' Car frames (wheels turned right)

	Field engine:Sound			' Engine sound
	Field crash:Sound				' Crash sound
	
	Field engineticks:Int			' Used for crudely timing engine sound length
	
	Field dotx:Float [numdots]		' Array of dot x-positions
	Field doty:Float [numdots]		' Array of dot x-positions
	Field dotsize:Float [numdots]	' Array of dot sizes

	Field credit:Image			' Reiner image credit
		
	Field frame:Int				' Current frame number
	Field updatecounter:Int		' Counts application updates
	
	Field x:Float				' Player's x position
	Field y:Float				' Player's y position
	
	Field speed:Float				' Player's speed
	Field acc:Float = 0.1			' Acceleration value
	Field angle:Float				' Player's angle

	Field dw:Float				' VDeviceWidth (gets cast to float)
	Field dh:Float				' VDeviceHeight (gets cast to float)
	
#rem
	Method OnLoading ()
	
		Local ticks:Int = Millisecs
		Local tx, ty
		
		If Millisecs > ticks + 1000
			ticks = Millisecs
			tx = Rnd (DeviceWidth)
			ty = Rnd (DeviceHeight)
		Endif
		
		DrawText "Still loading...", tx, ty
		
	End
#end

	Method OnCreate ()		' Load media and SET UPDATE RATE here...

		SetVirtualDisplay 800, 534
		
		' Store VDeviceWidth/Height as floats...
		
		dw = VDeviceWidth
		dh = VDeviceHeight

		' Load sounds...
		
		engine = LoadSound ("engine.wav")
		crash = LoadSound ("crash.wav")
		
		' Initialise ground detail...
		
		For Local loop:Int = 0 Until numdots
			dotx [loop] = Rnd (dw)
			doty [loop] = Rnd (dh)
			dotsize [loop] = Rnd (1, 4)
		Next
	
		' Build filenames and load frames...
		
		Local num:String ' String re-used to build filenames...

		Local frame:Int
		
		For frame = 0 To 31
			num = PadLeftZero (frame, 4) ' Pads frame out to 0001, 0002, etc...
			carstraight [frame] = LoadCenteredImage ("straight" + num + ".png")
		Next
		
		For frame = 0 To 31
			num = PadLeftZero (frame, 4)
			carleft [frame] = LoadCenteredImage ("turns left" + num + ".png")
		Next
		
		For frame = 0 To 31
			num = PadLeftZero (frame, 4)
			carright [frame] = LoadCenteredImage ("turns right" + num + ".png")
		Next

		' Load credit image...
		
		credit = LoadCenteredImage ("credit.png")
		
		' Set player to centre of screen...
		
		x = dw * 0.5
		y = dh * 0.5

		' 60 frames per second, driver!
		
		SetUpdateRate 60
		
		' Engine timer/start playing...

		engineticks = Millisecs
		
		' Set start frame...
		
		useframe = carstraight [0]

		' Ready to rock!
		
	End
	
	Method OnUpdate ()		' Take input and update game here...

		' Check engine timer, play sound if done...
			
		If Millisecs > engineticks + 1800
			engineticks = Millisecs
			' PlaySound engine ' AWAITING BUG FIX!
		Endif
		
		' Get these again each frame in case device resolution changes...
	
		dw = VDeviceWidth
		dh = VDeviceHeight

		' Keep track of updates (60 FPS)...
		
		updatecounter = updatecounter + 1
		If updatecounter > 59 Then updatecounter = 0

		' Speed...
		
		If KeyDown (KEY_UP)
			If speed < 4.0 Then speed = speed + acc ' Limited speed as 32 frames starts to look jerky when going fast!
		Else
			If KeyDown (KEY_DOWN)
				If speed > -2.0 Then speed = speed - acc ' Reverse
			Else
				speed = Diminish (speed, 0.02)
			Endif
		Endif

		' Steering...
		
		If KeyDown (KEY_LEFT)

			' Limited turning to every two frames, maximum, dependent also on speed (rate of turn)...
			
			If updatecounter Mod (Abs (speed) + 2) = 0	' Total trial and error! Still not good at very slow speeds
			
				If Sgn (speed) = 1
					frame = frame - 1
					If frame < 0 Then frame = 31		' ^
				Else							' This stuff checks for forward/reverse, and turns correctly
					If Sgn (speed) = -1			' v
						frame = frame + 1
						If frame > 31 Then frame = 0
					Endif
				Endif
				
			Endif
			
			' Set frame to be drawn...
			
			useframe = carleft [frame]

		Else
		
			' Same as for left turn, above...
			
			If KeyDown (KEY_RIGHT)

				If updatecounter Mod (Abs (speed) + 2) = 0
				
					If Sgn (speed) = 1
						frame = frame + 1
						If frame > 31 Then frame = 0
					Else
						If Sgn (speed) = -1
							frame = frame - 1
							If frame < 0 Then frame = 31
						Endif
					Endif
					
				Endif

				useframe = carright [frame]

			Else

				' Neither left nor right held down, so not steering...
			
				useframe = carstraight [frame]

			Endif

		Endif

		' Handbrake...
		
		If KeyDown (KEY_SPACE)
			speed = speed * 0.9
			If speed < 0.05 Then speed = 0.0
		Endif
		
		' Determine angle based on frame. Divide current frame by number of frames (32 for each wheel setting),
		' to get fraction; multiply that fraction by 360...
		
		angle = (Float (frame) / 32.0) * 360
		
		' SOH CAH TOA! Look it up. Gets horizontal/vertical movement based on angle/speed...
		
		Local opp:Float = Sin (angle) * speed
		Local adj:Float = Cos (angle) * speed
		
		x = x - opp ' Have to reverse this as car front is offset 180 degrees...
		y = y + adj ' ... or something. Dunno why this one stays positive. Hey ho...
		
		' Crash into borders...
		
		If x < 0
			x = 0
			speed = 0
			PlaySound crash
		Else
			If x > dw - 1
				x = dw - 1
				speed = 0
				PlaySound crash
			Endif
		Endif

		If y < 0
			y = 0
			speed = 0
			PlaySound crash
		Else
			If y > dh - 1
				y = dh - 1
				speed = 0
				PlaySound crash
			Endif
		Endif
		
	End
	
	Method OnRender ()		' Draw frame here...

		UpdateVirtualDisplay
		
		' Create a green and pleasant land...
		
'		Cls 32, 128, 64

		Cls 128, 112, 96
		
		' Going to be scaling, so store display matrix (see matrixrocks demo for more)...
		
		PushMatrix
		
			' Move drawing origin to centre of screen...
			
			Translate dw * 0.5, dh * 0.5

			' Get max distance from centre of display. Sqrt ((width * width) + (height * height)) gets diagonal length...
			
			Local halfdiag:Float = Sqrt ((dw * dw) + (dh * dh)) * 0.5

			' Distance from centre, divided by half of diagonal, gives value from 0.0 to 1.0 from centre out to corners.
			' This value is then reversed by taking it away from 1.0, making value from 1.0 to 0.0 instead...
			
			Local scalefactor:Float = 1.0 - (Distance (x, y, dw * 0.5, dh * 0.5) / halfdiag)
			
			 ' Adding 0.5 to scale, as it's 0 at the centre, which would draw nothing...
			
			Scale scalefactor + 0.5, scalefactor + 0.5
			
			' Shift origin back where it was, ie. 0, 0...
			
			Translate -dw * 0.5, -dh * 0.5

			' Draw border...
			
			SetColor 72, 64, 64
			
			DrawLine 0, 0, dw, 0
			DrawLine 0, 0, 0, dh
			DrawLine dw, 0, dw, dh
			DrawLine 0, dh, dw, dh

			' Draw ground detail...
			
			SetColor 132, 124, 116
			
			For Local loop:Int = 0 Until 100
				DrawRect dotx [loop], doty [loop], dotsize [loop], dotsize [loop]
			Next
			
			' Draw car...
			
			DrawImage useframe, x, y
		
		PopMatrix
		
		' Back to default scaling, so draw image credits...
		
		DrawImage credit, dw * 0.5, dh - credit.Height

	End
	
End

' Main function...

Function Main ()
	New Game
End

' MISC FUNCTIONS...

' Distance...

Function Distance:Float (x1:Float, y1:Float, x2:Float, y2:Float)
	Return Sqrt ((x1 - x2) * (x1 - x2) + (y1 - y2) * (y1 - y2))
End

' Set image's handle to centre...

Function MidHandle (image:Image)
	image.SetHandle image.Width () * 0.5, image.Height () * 0.5
End

' Load an image with handle already centered (uses MidHandle function above)...

Function LoadCenteredImage:Image (image:String)
	Local img:Image = LoadImage (image)
	MidHandle img
	Return img
End

' Quick helper function to add zeroes to left of number in string form...

Function PadLeftZero:String (num:String, size:Int)
	While num.Length < size
		num = "0" + num
	Wend
	Return num
End

Function Diminish:Float (value:Float, amount:Float)
	If Abs (value) > Abs (amount)
		Return value - (amount * Sgn (value))
	EndIf
End Function
